from dataclasses import dataclass
from enum import Enum
from functools import cached_property
import hashlib
import io
import json
import os
from pathlib import Path
import random
import sys
import time
import traceback
from typing import Any, List, Tuple
from typing import Iterator
from typing import Optional
from typing import Union

import lymbo
from lymbo.env import LYMBO_REPORT_PATH
from lymbo.env import LYMBO_TEST_SCOPE_CLASS
from lymbo.env import LYMBO_TEST_SCOPE_FUNCTION
from lymbo.env import LYMBO_TEST_SCOPE_MAX
from lymbo.env import LYMBO_TEST_SCOPE_MODULE
from lymbo.env import LYMBO_TEST_SCOPE_GLOBAL
from lymbo import color


class TestStatus(Enum):
    PENDING = "pending"
    INPROGRESS = "inprogress"
    PASSED = "passed"
    FAILED = "failed"
    BROKEN = "broken"
    SKIPPED = "skipped"


class ReportFailure(Enum):
    NONE = "none"
    SIMPLE = "simple"
    NORMAL = "normal"
    FULL = "full"


class GroupBy(Enum):
    NONE = "none"
    MODULE = "module"
    CLASS = "class"
    FUNCTION = "function"


class TestItem:
    """A test"""

    def __init__(
        self,
        path: Path,
        asynchronous: bool,
        fnc: str,
        parameters: tuple[tuple[Any], dict[str, Any]],
        cls: Optional[str],
    ):
        self.path = path
        self.asynchronous = asynchronous
        self.fnc = fnc
        self.parameters = parameters
        self.cls = cls

        md5 = hashlib.md5(str(self).encode()).hexdigest()
        timestamp = int(time.time() * 1000000)
        rnd = random.randint(0, 99999)
        self.uuid = f"{md5}-{timestamp}-{rnd:05d}"

        self.start_at: float = 0.0
        self.end_at: float = 0.0

        self.output: io.StringIO = io.StringIO()

        self.status: TestStatus = TestStatus.PENDING
        self.reason: str = ""
        self.error_message: List[str] = []
        self.traceback: List[str] = []

    def __str__(self):

        def print_variable(variable):
            if isinstance(variable, str):
                return f'"{variable}"'
            else:
                return str(variable)

        s = f"{self.path}::"
        if self.cls:
            s += f"{self.cls}::"
        s += f"{self.fnc}"
        args, kwargs = self.parameters
        s += "("
        call = []
        for arg in args:
            call.append(str(arg))
        for k, v in kwargs.items():
            call.append(f"{k}={print_variable(v)}")
        s += ",".join(call)
        s += ")"

        return s

    def __repr__(self) -> str:
        return f"{self.uuid}%{self}"

    def to_json(self):
        return {
            "lymbo": lymbo.__version__,
            "test": {
                "name": str(self),
                "uuid": self.uuid,
                "status": self.status.value,
                "start_at": self.start_at,
                "end_at": self.end_at,
                "output": self.output.getvalue(),
                "error": {
                    "reason": self.reason,
                    "error_message": self.error_message,
                    "traceback": self.traceback,
                },
            },
        }

    def write_report(self):

        path = f"{os.environ[LYMBO_REPORT_PATH]}/lymbo-{self.uuid}"

        with open(path + ".tmp", "w") as f:
            json.dump(self.to_json(), f, indent=4)

        os.replace(path + ".tmp", path + ".json")

    def refresh_from_report(self):

        path = f"{os.environ[LYMBO_REPORT_PATH]}/lymbo-{self.uuid}.json"

        if os.path.exists(path):

            with open(path) as f:
                test_desc = json.load(f)["test"]

                assert test_desc["uuid"] == self.uuid

                self.status = TestStatus(test_desc["status"])
                self.start_at = test_desc["start_at"]
                self.end_at = test_desc["end_at"]
                self.output = io.StringIO(test_desc["output"])
                self.reason = test_desc["error"]["reason"]
                self.error_message = test_desc["error"]["error_message"]
                self.traceback = test_desc["error"]["traceback"]

    def start(self):
        self.start_at = time.time()
        sys.stdout = self.output
        sys.stderr = self.output
        self.write_report()

    def end(self, reason: Union[Exception, None] = None):
        self.end_at = time.time()
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        self.reason = str(reason)
        if reason is None:
            self.status = TestStatus.PASSED
        elif isinstance(reason, AssertionError):
            self.status = TestStatus.FAILED
            if not self.reason:
                self.reason = "AssertionError (no message)"
        else:
            self.status = TestStatus.BROKEN
        self.error_message = self.__error_message(reason)

        self.traceback = self.__traceback(reason)

        self.write_report()

    @property
    def duration(self):
        return self.end_at - self.start_at

    @cached_property
    def scopes(self) -> dict:
        scopes = {
            LYMBO_TEST_SCOPE_MODULE: f"{self.path}",
            LYMBO_TEST_SCOPE_GLOBAL: LYMBO_TEST_SCOPE_GLOBAL,
            LYMBO_TEST_SCOPE_MAX: "global",
        }

        if self.cls:
            scopes[LYMBO_TEST_SCOPE_CLASS] = f"{self.path}::{self.cls}"
            scopes[LYMBO_TEST_SCOPE_FUNCTION] = f"{self.path}::{self.cls}::{self.fnc}"
        else:
            scopes[LYMBO_TEST_SCOPE_FUNCTION] = f"{self.path}::{self.fnc}"
            scopes[LYMBO_TEST_SCOPE_CLASS] = scopes[LYMBO_TEST_SCOPE_FUNCTION]

        return scopes

    def __error_message(self, reason) -> List[str]:

        message = []

        first_line_test = 3 if self.asynchronous else 2

        if reason:
            tb = traceback.extract_tb(reason.__traceback__)

            if len(tb) > first_line_test:
                # We get the line in the test where the exception has been raised
                filename, lineno, funcname, text = tb[first_line_test]

                message.append(
                    f"{type(reason).__name__} in {filename}, line {lineno}, in {funcname}:"
                )

                # To have a better context, we retrieve 2 lines before the error, and 1 line after
                with open(filename, "r") as f:
                    lines = f.readlines()

                start_line = max(0, lineno - 3)
                end_line = min(lineno + 1, len(lines) - 1)

                for i in range(start_line, end_line + 1):
                    flag = (
                        "==> " if i == lineno - 1 else "    "
                    )  # to indicate where was the error
                    message.append(f"{flag}{i + 1}: {lines[i].rstrip()}")
            else:
                # the exception occured before the execution of the test
                message.append(
                    f"The test has not been executed - [{type(reason).__name__}: {str(reason)}]"
                )

        return message

    @staticmethod
    def __traceback(reason) -> List[str]:

        trcbck = []

        if reason:

            for lines in traceback.format_exception(
                type(reason), reason, reason.__traceback__
            )[3:]:
                for line in lines.split("\n"):
                    trcbck.append(line)

        return trcbck


@dataclass
class TestPlan:
    groups: list[list[TestItem]]
    group_by: GroupBy

    def __iter__(self) -> Iterator[list[TestItem]]:
        for group in self.groups:
            yield group

    @cached_property
    def count(self) -> tuple[int, int]:
        """Number of tests and number of groups"""
        nb_groups = len(self.groups)
        nb_tests = 0

        for group in self.groups:
            nb_tests += len(group)

        return nb_tests, nb_groups

    def __str__(self) -> str:
        test_plan, _ = self.test_plan()
        return test_plan

    def test_plan(
        self,
        show_status: bool = False,
    ) -> Tuple[str, dict[TestStatus, int]]:
        output: list[str] = []
        tests_status = {status: 0 for status in TestStatus}
        for tests in self.groups:
            if len(tests) > 1:
                group_msg = f"+ {len(tests)} tests grouped by "
                if self.group_by == GroupBy.MODULE:
                    group_msg += f"{tests[0].path}"
                if self.group_by == GroupBy.CLASS:
                    group_msg += f"{tests[0].path}::{tests[0].cls}"
                if self.group_by == GroupBy.FUNCTION:
                    group_msg += f"{tests[0].path}::{tests[0].cls}::{tests[0].fnc}"
                output.append(group_msg)
            for test in tests:
                repr = f"{'  | -' if len(tests)>1 else '-'} {test}"
                if show_status:
                    test.refresh_from_report()
                    tests_status[test.status] += 1
                    color_per_status = {
                        TestStatus.PENDING: color.RESET,
                        TestStatus.INPROGRESS: color.RESET,
                        TestStatus.PASSED: color.GREEN,
                        TestStatus.FAILED: color.RED,
                        TestStatus.BROKEN: color.YELLOW,
                        TestStatus.SKIPPED: color.RESET,
                    }
                    repr += f" [{color_per_status[test.status]}{test.status.value.upper()}{color.RESET}]"
                output.append(repr)

        return "\n".join(output), tests_status

    def failures(
        self,
        report_failure: ReportFailure = ReportFailure.NORMAL,
    ) -> Tuple[str, int]:
        output: list[str] = []

        nb = 0
        for tests in self.groups:
            for test in tests:
                test.refresh_from_report()
                if test.status in (TestStatus.BROKEN, TestStatus.FAILED):
                    nb += 1
                    padding = "  "
                    output.append(str(test))
                    test_output = [line for line in test.output.getvalue().split("\n")]
                    if test_output and not test_output[-1].strip():
                        del test_output[-1]
                    if test_output:
                        output.append(" - - - - output:")
                        for line in test_output:
                            output.append(f"{padding}{line}")
                    output.append(" - - - - reason:")
                    if report_failure in (
                        ReportFailure.SIMPLE,
                        ReportFailure.NORMAL,
                        ReportFailure.FULL,
                    ):
                        output.append(f"{padding}{test.reason}")
                    if report_failure in (ReportFailure.NORMAL, ReportFailure.FULL):
                        output.append(f"{padding}---------------------")
                        for line in test.error_message:
                            not_empty_line = line.rstrip()
                            if not_empty_line:
                                output.append(f"{padding}{not_empty_line}")
                    if report_failure == ReportFailure.FULL:
                        output.append(f"{padding}---------------------")
                        for line in test.traceback:
                            not_empty_line = line.rstrip()
                            if not_empty_line:
                                output.append(f"{padding}{not_empty_line}")
        return "\n".join(output), nb



import unittest.mock
from lymbo import args
from lymbo import expand
from lymbo.env import LYMBO_TEST_COLLECTION

import os
import unittest


class TestArgs(unittest.TestCase):

    @unittest.mock.patch.dict(os.environ, {LYMBO_TEST_COLLECTION: "1"})
    def test_args_default(self):

        flattened_args = args()

        self.assertEqual(flattened_args, [((), {})])

    @unittest.mock.patch.dict(os.environ, {LYMBO_TEST_COLLECTION: "1"})
    def test_args_no_param(self):

        flattened_args = args(123, "hello", name="opt")

        self.assertEqual(flattened_args, [((123, "hello"), {"name": "opt"})])

    @unittest.mock.patch.dict(os.environ, {LYMBO_TEST_COLLECTION: "1"})
    def test_args_no_param_list(self):

        flattened_args = args(123, ["hello", "salut"], name=["opt", "glo"])

        self.assertEqual(
            flattened_args,
            [((123, ["hello", "salut"]), {"name": ["opt", "glo"]})],
        )

    @unittest.mock.patch.dict(os.environ, {LYMBO_TEST_COLLECTION: "1"})
    def test_args_one_param_arg(self):

        flattened_args = args(123, expand("hello", "salut"), name="opt")

        self.assertEqual(
            flattened_args,
            [((123, "hello"), {"name": "opt"}), ((123, "salut"), {"name": "opt"})],
        )

    @unittest.mock.patch.dict(os.environ, {LYMBO_TEST_COLLECTION: "1"})
    def test_args_one_param_kwarg(self):

        flattened_args = args(123, "salut", name=expand("opt", "glo"))

        self.assertEqual(
            flattened_args,
            [((123, "salut"), {"name": "opt"}), ((123, "salut"), {"name": "glo"})],
        )

    @unittest.mock.patch.dict(os.environ, {LYMBO_TEST_COLLECTION: "1"})
    def test_args_many_params(self):

        flattened_args = args(
            123, expand("salut", "hello", "ciao"), name=expand("opt", "glo")
        )

        self.assertCountEqual(
            flattened_args,
            [
                ((123, "salut"), {"name": "opt"}),
                ((123, "salut"), {"name": "glo"}),
                ((123, "hello"), {"name": "opt"}),
                ((123, "hello"), {"name": "glo"}),
                ((123, "ciao"), {"name": "opt"}),
                ((123, "ciao"), {"name": "glo"}),
            ],
        )

    def test_args_not_in_collect(self):

        self.assertIsNone(
            args(123, expand("salut", "hello", "ciao"), name=expand("opt", "glo"))
        )


if __name__ == "__main__":
    unittest.main()



import os
from pathlib import Path
import unittest

from lymbo.collect import collect_tests
from lymbo.collect import extract_words_from_filter
from lymbo.collect import match_filter
from lymbo.exception import LymboExceptionFilter
from lymbo.item import GroupBy

dir = os.path.dirname(os.path.abspath(__file__))


class TestCollect(unittest.TestCase):

    # group by

    def test_collect_tests_groupby_none(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect"))], GroupBy.NONE
        )

        self.assertEqual(test_plan.count, (11, 11))

    def test_collect_tests_groupby_function(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect"))], GroupBy.FUNCTION
        )

        self.assertEqual(test_plan.count, (11, 6))

    def test_collect_tests_groupby_class(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect"))], GroupBy.CLASS
        )

        self.assertEqual(test_plan.count, (11, 10))

    def test_collect_tests_groupby_module(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect"))], GroupBy.MODULE
        )

        self.assertEqual(test_plan.count, (11, 2))

    # list files

    def test_collect_tests_dir_not_exists(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect_xxxxx"))], GroupBy.NONE
        )

        self.assertEqual(test_plan.count, (0, 0))

    def test_collect_tests_dir(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect"))], GroupBy.NONE
        )

        self.assertEqual(test_plan.count, (11, 11))

    def test_collect_tests_one_file(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect/collect_a.py"))], GroupBy.NONE
        )

        self.assertEqual(test_plan.count, (4, 4))

    def test_collect_tests_two_files(self):

        test_plan = collect_tests(
            [
                Path(os.path.join(dir, "data_collect/collect_a.py")),
                Path(os.path.join(dir, "data_collect/collect_b.py")),
            ],
            GroupBy.NONE,
        )

        self.assertEqual(test_plan.count, (11, 11))

    # filters

    def test_extract_words(self):

        params = {
            "abc": ["abc"],
            "not abc": ["abc"],
            "not (abc)": ["abc"],
            "(not abc)": ["abc"],
            "abc or abc": ["abc"],
            "abc or abcdef": ["abc", "abcdef"],
            "abc or def": ["abc", "def"],
            "abc and def": ["abc", "def"],
            "abc or (def and ijk)": ["abc", "def", "ijk"],
            "abc or not (def and ijk)": ["abc", "def", "ijk"],
            "(  abc or ((def and (ijk))    )": ["abc", "def", "ijk"],
        }

        for filter, words in params.items():
            with self.subTest(f'filter="{filter}" words={words}'):
                self.assertListEqual(
                    sorted(extract_words_from_filter(filter)), sorted(words)
                )

    def test_match_filter(self):

        params = {
            "abc": "abc",
            "abc/def/ijk.py::func1": "abc",
            "abc/def/ijk.py::func2": "abc or abcdef",
            "abc/def/ijk.py::func3": "abcdef or abc",
            "abc/def/ijk.py::func4": "def",
            "abc/def/ijk.py::func5": "abc and def",
            "abc/def/ijk.py::func6": "abc and def and not ABC",
        }

        for item, filter in params.items():
            with self.subTest(f'item="{item}" filter="{filter}"'):
                self.assertTrue(match_filter(item, filter))

    def test_match_filter_not_match(self):

        params = {
            "abc": "def",
            "abc/def/ijk.py::func1": "ABC",
            "abc/def/ijk.py::func2": "abc and abcdef",
            "abc/def/ijk.py::func3": "abcdef and abc",
            "abc/def/ijk.py::func4": "not def",
            "abc/def/ijk.py::func5": "abc and not def",
            "abc/def/ijk.py::func6": "not abc or (DEF and not ABC)",
        }

        for item, filter in params.items():
            with self.subTest(f'item="{item}" filter="{filter}"'):
                self.assertFalse(match_filter(item, filter))

    def test_collect_tests_filter(self):

        test_plan = collect_tests(
            [Path(os.path.join(dir, "data_collect"))],
            GroupBy.MODULE,
            "second and not ((p=4) or (p=5))",
        )

        self.assertEqual(test_plan.count, (5, 2))

    def test_collect_tests_broken(self):

        try:
            _ = collect_tests(
                [Path(os.path.join(dir, "data_collect"))], GroupBy.MODULE, "second )"
            )
            self.assertFalse(True, "No exception has been raised.")
        except Exception as ex:
            self.assertIsInstance(ex, LymboExceptionFilter)


if __name__ == "__main__":
    unittest.main()



from lymbo.cm import args
from lymbo.cm import expand
from lymbo.cm import test
from lymbo.resource_manager import scope_class
from lymbo.resource_manager import scope_function
from lymbo.resource_manager import scope_global
from lymbo.resource_manager import scope_module

__version__ = "0.1.0"

__all__ = [
    "args",
    "expand",
    "test",
    "scope_class",
    "scope_function",
    "scope_global",
    "scope_module",
]



LYMBO_REPORT_PATH = "LYMBO_REPORT_PATH"

LYMBO_TEST_COLLECTION = "LYMBO_TEST_COLLECTION"

LYMBO_TEST_SCOPE_CLASS = "LYMBO_TEST_SCOPE_CLASS"
LYMBO_TEST_SCOPE_FUNCTION = "LYMBO_TEST_SCOPE_FUNCTION"
LYMBO_TEST_SCOPE_MODULE = "LYMBO_TEST_SCOPE_MODULE"
LYMBO_TEST_SCOPE_GLOBAL = "LYMBO_TEST_SCOPE_GLOBAL"
LYMBO_TEST_SCOPE_MAX = "LYMBO_TEST_SCOPE_MAX"

LYMBO_LOG_LEVEL = "LYMBO_LOG_LEVEL"
LYMBO_LOG_PATH = "LYMBO_LOG_PATH"



class LymboExceptionScopeHierarchy(Exception):
    """A shared resource is initialized in a higher scope than allowed."""

    pass


class LymboExceptionScopeNested(Exception):
    """A shared resource is initialized from another shared resource."""

    pass


class LymboExceptionFilter(Exception):
    """The test collection filter is broken."""

    pass



import ast
import glob
import importlib
import os
from pathlib import Path
import sys
from typing import Union
from unittest.mock import patch

from lymbo.config import GroupBy
from lymbo.env import LYMBO_TEST_COLLECTION
from lymbo.exception import LymboExceptionFilter
from lymbo.item import TestItem
from lymbo.item import TestPlan
from lymbo.log import trace_call
from lymbo.log import logger


from lymbo.cm import args


@trace_call
def collect_tests(
    paths: list[Path], group_by: GroupBy, filter_by_path: str = ""
) -> TestPlan:
    """Collect all the functions/methods decorated with @lymbo.test."""

    tests = []
    filtered_tests = []

    os.environ[LYMBO_TEST_COLLECTION] = "1"

    for path in list_python_files(paths):

        tests += list_tests_from_file(path, group_by)

    del os.environ[LYMBO_TEST_COLLECTION]

    if filter_by_path:
        for group in tests:
            new_group = []
            for test in group:
                if match_filter(str(test), filter_by_path):
                    new_group.append(test)
                else:
                    logger().debug(f"collect_tests - {str(test)} has been filtered.")

            if new_group:
                filtered_tests.append(new_group)
    else:
        filtered_tests = tests

    test_plan = TestPlan(filtered_tests, group_by)

    return test_plan


## List tests


@trace_call
def list_python_files(paths: list[Path]) -> list[Path]:
    """Walk into all directories and subdirectories to list the Python files."""
    tests_files = []
    for path in paths:
        if path.is_file() and path.name.endswith(".py"):
            tests_files.append(path)
        elif path.is_dir():
            if not path.name.startswith("__"):
                for p in glob.glob(f"{path}/**"):
                    tests_files += list_python_files(
                        [
                            Path(p),
                        ]
                    )
    return list(set(tests_files))


@trace_call
def list_tests_from_file(path: Path, group_by: GroupBy) -> list[list[TestItem]]:
    """List all the tests defined in a file."""

    collected_tests = []

    with open(path) as f:
        source = f.read()

    syspath = sys.path + [
        str(path.parent.absolute()),
    ]

    with patch.object(sys, "path", syspath):

        tree = ast.parse(source, path)

        # get the context
        imports = extract_imports(tree)
        global_vars = dynamic_import_modules(
            imports
        )  # we must ensure the import are done, if the module contains a class

        local_vars: dict[str, str] = {}
        compiled_code = compile(source, path, "exec")
        exec(
            compiled_code, global_vars, local_vars
        )  # we execute the module to retrieve the global and local vars

        tests = parse_body(group_by, tree.body, path, None, global_vars, local_vars)
        if group_by == GroupBy.MODULE:
            collected_tests.append([test[0] for test in tests])
        else:
            collected_tests += tests

    return collected_tests


def parse_body(
    group_by: GroupBy,
    body: list[ast.stmt],
    path: Path,
    classdef: Union[ast.ClassDef, None],
    global_vars,
    local_vars,
) -> list[list[TestItem]]:
    """Parse the body a module/class to find test."""
    collected_tests = []
    for item in body:
        if isinstance(item, (ast.FunctionDef, ast.AsyncFunctionDef)):
            for decorator in item.decorator_list:
                if isinstance(decorator, ast.Call):
                    if isinstance(decorator.func, ast.Attribute):
                        if decorator.func.attr == "test":
                            if getattr(decorator.func.value, "id", None) == "lymbo":

                                args_call = None

                                if decorator.args:
                                    args_call = decorator.args[0]
                                else:
                                    for kw in decorator.keywords:
                                        if kw.arg == "args":
                                            args_call = kw.value

                                if args_call:
                                    flattened_args = eval_ast_call(
                                        args_call, global_vars, local_vars
                                    )
                                else:
                                    flattened_args = args()

                                tests = []
                                for f_args in flattened_args:
                                    tests.append(
                                        [
                                            TestItem(
                                                path,
                                                isinstance(item, ast.AsyncFunctionDef),
                                                item.name,
                                                f_args,
                                                classdef.name if classdef else None,
                                            ),
                                        ]
                                    )
                                if (len(tests) > 1) and (group_by == GroupBy.FUNCTION):
                                    collected_tests.append([test[0] for test in tests])
                                else:
                                    # error: Argument 1 to "extend" of "list" has
                                    # incompatible type "list[list[TestItem]]"; expected
                                    # "Iterable[list[list[TestItem]]]"
                                    collected_tests.extend(tests)  # type: ignore[arg-type]
        elif isinstance(item, ast.ClassDef):
            tests = parse_body(group_by, item.body, path, item, global_vars, local_vars)
            if group_by == GroupBy.CLASS:
                collected_tests.append([test[0] for test in tests])
            else:
                # error: Argument 1 to "extend" of "list" has incompatible
                # type "list[list[TestItem]]"; expected "Iterable[list[list[TestItem]]]"
                collected_tests.extend(tests)  # type: ignore[arg-type]

    return collected_tests


def eval_ast_call(call_node, global_vars, local_vars):

    # local_vars and global_vars should be passed in the context
    # where the AST was created

    # Retrieve the function object
    func = eval(
        compile(ast.Expression(call_node.func), "", mode="eval"),
        global_vars,
        local_vars,
    )

    # Evaluate the arguments (handling more complex expressions, not just literals)
    args = [
        eval(compile(ast.Expression(arg), "", mode="eval"), global_vars, local_vars)
        for arg in call_node.args
    ]

    # Evaluate the keyword arguments (handling more complex expressions)
    kwargs = {
        kw.arg: eval(
            compile(ast.Expression(kw.value), "", mode="eval"), global_vars, local_vars
        )
        for kw in call_node.keywords
    }

    # Call the function with the evaluated arguments and keyword arguments
    return func(*args, **kwargs)


def set_parent_references(node: ast.AST, parent: ast.AST = None):
    """Set parent references in all nodes."""
    node.parent = parent  # type: ignore[attr-defined]
    for child in ast.iter_child_nodes(node):
        set_parent_references(child, node)


def extract_imports(tree: ast.Module) -> list[tuple[str, str]]:
    """Extracts all top-level imports from the given module and returns them
    as a list of module names or alias."""

    set_parent_references(tree)  # to detect if this is a top level import or not

    imports = []
    for node in ast.walk(tree):
        if isinstance(node, ast.Import) and isinstance(
            node.parent, (ast.Module, ast.ClassDef)  # type: ignore[attr-defined]
        ):
            for alias in node.names:
                imports.append(
                    (alias.name, alias.asname if alias.asname else alias.name)
                )
        elif isinstance(node, ast.ImportFrom) and isinstance(
            node.parent, (ast.Module, ast.ClassDef)  # type: ignore[attr-defined]
        ):
            module_name = node.module
            for alias in node.names:
                full_name = f"{module_name}.{alias.name}"
                imports.append(
                    (full_name, alias.asname if alias.asname else alias.name)
                )

    return imports


def dynamic_import_modules(imports: list[tuple[str, str]]) -> dict[str, str]:
    """Dynamically imports the modules and returns a dictionary of imported names."""
    global_vars = {}

    for full_name, alias in imports:
        # Import the module or specific attribute
        module_name, _, attr_name = full_name.rpartition(".")
        if module_name:  # This handles `from module import name`
            module = importlib.import_module(module_name)
            try:
                # Try to get the attribute from the module
                global_vars[alias] = getattr(module, attr_name)
            except AttributeError:
                # If it's not an attribute, treat it as a submodule and import it
                global_vars[alias] = importlib.import_module(full_name)
        else:  # This handles `import module`
            global_vars[alias] = importlib.import_module(full_name)

    return global_vars


## Filter tests


def extract_words_from_filter(filter: str) -> list[str]:
    """Parse a filter and returns only the words of the expression.
    Example: filter = "abc and not (def or ghi)"
             output = ["abc", "def", "ghi"]
    """

    # remove the stop words
    stop_words = ["(", ")", "and", "or", "not"]
    for stop_word in stop_words:
        filter = filter.replace(stop_word, " ")

    # extract words
    words = set(filter.split(" "))

    return [word for word in words if word != ""]


def match_filter(item: str, filter: str) -> bool:
    """Indicate if this item match with the filter"""
    words = {word: False for word in extract_words_from_filter(filter)}

    for word in words:
        words[word] = word in item

    # We sort the list of words to first replace the longer words to have partial
    # replacement if for example we have these two words: abc and abcdef.
    for word in sorted(words, key=len, reverse=True):
        filter = filter.replace(word, str(words[word]))

    try:
        match = eval(filter)
    except Exception as ex:
        raise LymboExceptionFilter(
            f'The filter ["{filter}"] is broken. exception=[{str(ex)}]'
        )

    return match
